@@ -1789,7 +1789,7 @@ find_file_in_data_dirs (const gchar   *file,
     {
       g_set_error (error, G_BOOKMARK_FILE_ERROR,
                    G_BOOKMARK_FILE_ERROR_FILE_NOT_FOUND,
-                   _("No valid bookmark file was be found in data dirs"));
+                   _("No valid bookmark file found in data dirs"));
       
       return NULL;
     }
@@ -2016,17 +2016,19 @@ g_bookmark_file_add_item (GBookmarkFile  *bookmark,
  *
  * Removes the bookmark for @uri from the bookmark file @bookmark.
  *
+ * Return value: %TRUE if the bookmark was removed successfully.
+ * 
  * Since: 2.12
  */
-void
+gboolean
 g_bookmark_file_remove_item (GBookmarkFile  *bookmark,
 			     const gchar    *uri,
 			     GError        **error)
 {
   BookmarkItem *item;
   
-  g_return_if_fail (bookmark != NULL);
-  g_return_if_fail (uri != NULL);
+  g_return_val_if_fail (bookmark != NULL, FALSE);
+  g_return_val_if_fail (uri != NULL, FALSE);
   
   item = g_bookmark_file_lookup_item (bookmark, uri);
   
@@ -2036,13 +2038,15 @@ g_bookmark_file_remove_item (GBookmarkFile  *bookmark,
 		   G_BOOKMARK_FILE_ERROR_URI_NOT_FOUND,
 		   _("No bookmark found for URI '%s'"),
 		   uri);
-      return;	 
+      return FALSE;
     }
 
   bookmark->items = g_list_remove (bookmark->items, item);
   g_hash_table_remove (bookmark->items_by_uri, item->uri);  
   
   bookmark_item_free (item);
+
+  return TRUE;
 }
 
 /**
@@ -2838,7 +2842,8 @@ g_bookmark_file_remove_group (GBookmarkFile  *bookmark,
       if (strcmp (l->data, group) == 0)
         {
           item->metadata->groups = g_list_remove_link (item->metadata->groups, l);
-          g_list_free_1 (l);
+          g_free (l->data);
+	  g_list_free_1 (l);
           
           item->modified = time (NULL);          
           
@@ -3244,7 +3249,8 @@ g_bookmark_file_set_app_info (GBookmarkFile  *bookmark,
     {
       item->metadata->applications = g_list_remove (item->metadata->applications, ai);
       g_hash_table_remove (item->metadata->apps_by_name, ai->name);
-          
+      bookmark_app_info_free (ai);
+
       item->modified = time (NULL);
           
       return TRUE;
